#include <iostream>
#include <stdio.h>
#include <stdlib.h>
#include <vector>
#include <fstream>

using std::cerr;
using std::endl;

#include <GL/glew.h>    // include GLEW and new version of GL on Windows
#include <GLFW/glfw3.h> // GLFW helper library

#define GLM_ENABLE_EXPERIMENTAL
#include <glm/vec3.hpp>   // glm::vec3
#include <glm/vec4.hpp>   // glm::vec4
#include <glm/mat4x4.hpp> // glm::mat4
#include <glm/gtx/string_cast.hpp>
#include <glm/gtc/type_ptr.hpp>
#include <glm/gtc/matrix_transform.hpp> // glm::translate, glm::rotate, glm::scale

//#define TINYOBJLOADER_IMPLEMENTATION
//#include "tiny_obj_loader.h"

#define STB_IMAGE_IMPLEMENTATION
#include "stb_image.h"

class RenderManager;

void SetUpDog(double, RenderManager &);
const char *GetVertexShader();
const char *GetFragmentShader();

uint8_t* load_images(char* path, int* w, int* h){
    //TODO: IMPLEMENT
    //LAYOUT: [r0, g0, b0, r1, g1, b1 ...]
    //the mapping provided by stb_image matches with what OpenGL expects
    //so there is no need to reorder the pixels in any way
    //SOME INPUT IMAGES WILL BE IN GRAYSCALE
    //IF YOU SET THE desired_channels/req_com (whatever they call the 4th argument) 
    //TO 3, stbi_load_from_file WILL GIVE YOU A BUFFER WITH THREE CHANNELS
    //AND MIRROR THE INPUT VALUE TO THE RED, GREEN, AND BLUE CHANNELS
    //THIS IS EXACTLY WHAT YOU WANT 


    //return mallocd_image_buffer;
}

GLuint load_opengl_image(char* path){
    //TODO
    //INSERT THE IMAGE IN TO OPENGL HERE (glGenTexture plus all the other stuff)
    //RETURN THE GLuint HANDLE TO THE IMAGE
    //THIS WILL GET INVOKED ONCE PER IMAGE
    int w;
    int h;

    uint8_t* tex = load_images(path, &w, &h);

    //return tex_gluint;
}

void setup_opengl_images(char* base_path, GLuint* alb_int, GLuint* rgh_int, GLuint* met_int){
    char alb_path[256];
    sprintf(alb_path, "%s%s", base_path, "_Color.jpg");
    char roughness_path[256];
    sprintf(roughness_path, "%s%s", base_path, "_Roughness.jpg");
    char metalness_path[256];
    sprintf(metalness_path, "%s%s", base_path, "_Metalness.jpg");

    *alb_int = load_opengl_image(alb_path);
    *rgh_int = load_opengl_image(roughness_path);
    *met_int = load_opengl_image(metalness_path);
}

//
//
// PART 1: code to set up spheres and cylinders
//
//

class Triangle
{
public:
    glm::vec3 v0;
    glm::vec3 v1;
    glm::vec3 v2;
};

// typedef struct sphere_s {
//     std::vector<float> verts;
//     std::vector<float> norms;
//     std::vector<float> uvs;
// } sphere_t;

std::vector<Triangle> SplitTriangle(std::vector<Triangle> &list)
{
    std::vector<Triangle> output(4 * list.size());
    output.resize(4 * list.size());
    for (unsigned int i = 0; i < list.size(); i++)
    {
        Triangle t = list[i];
        glm::vec3 vmid1, vmid2, vmid3;
        vmid1 = (t.v0 + t.v1) / 2.0f;
        vmid2 = (t.v1 + t.v2) / 2.0f;
        vmid3 = (t.v0 + t.v2) / 2.0f;
        output[4 * i + 0].v0 = t.v0;
        output[4 * i + 0].v1 = vmid1;
        output[4 * i + 0].v2 = vmid3;
        output[4 * i + 1].v0 = t.v1;
        output[4 * i + 1].v1 = vmid2;
        output[4 * i + 1].v2 = vmid1;
        output[4 * i + 2].v0 = t.v2;
        output[4 * i + 2].v1 = vmid3;
        output[4 * i + 2].v2 = vmid2;
        output[4 * i + 3].v0 = vmid1;
        output[4 * i + 3].v1 = vmid2;
        output[4 * i + 3].v2 = vmid3;
    }
    return output;
}

void PushVertex(std::vector<float> &coords,
                const glm::vec3 &v)
{
    coords.push_back(v.x);
    coords.push_back(v.y);
    coords.push_back(v.z);
}

void PushUV(std::vector<float>& uvs, const glm::vec2& uv){
    uvs.push_back(uv.x);
    uvs.push_back(uv.y);
    //uvs.push_back(1.0);
}

//
// Sets up a cylinder that is the circle x^2+y^2=1 extruded from
// Z=0 to Z=1.
//
void GetCylinderData(std::vector<float> &coords, std::vector<float> &normals)
{
    int nfacets = 30;
    for (int i = 0; i < nfacets; i++)
    {
        double angle = 3.14159 * 2.0 * i / nfacets;
        double nextAngle = (i == nfacets - 1 ? 0 : 3.14159 * 2.0 * (i + 1) / nfacets);
        glm::vec3 fnormal(0.0f, 0.0f, 1.0f);
        glm::vec3 bnormal(0.0f, 0.0f, -1.0f);
        glm::vec3 fv0(0.0f, 0.0f, 1.0f);
        glm::vec3 fv1(cos(angle), sin(angle), 1);
        glm::vec3 fv2(cos(nextAngle), sin(nextAngle), 1);
        glm::vec3 bv0(0.0f, 0.0f, 0.0f);
        glm::vec3 bv1(cos(angle), sin(angle), 0);
        glm::vec3 bv2(cos(nextAngle), sin(nextAngle), 0);
        // top and bottom circle vertices
        PushVertex(coords, fv0);
        PushVertex(normals, fnormal);
        PushVertex(coords, fv1);
        PushVertex(normals, fnormal);
        PushVertex(coords, fv2);
        PushVertex(normals, fnormal);
        PushVertex(coords, bv0);
        PushVertex(normals, bnormal);
        PushVertex(coords, bv1);
        PushVertex(normals, bnormal);
        PushVertex(coords, bv2);
        PushVertex(normals, bnormal);
        // curves surface vertices
        glm::vec3 v1normal(cos(angle), sin(angle), 0);
        glm::vec3 v2normal(cos(nextAngle), sin(nextAngle), 0);
        //fv1 fv2 bv1
        PushVertex(coords, fv1);
        PushVertex(normals, v1normal);
        PushVertex(coords, fv2);
        PushVertex(normals, v2normal);
        PushVertex(coords, bv1);
        PushVertex(normals, v1normal);
        //fv2 bv1 bv2
        PushVertex(coords, fv2);
        PushVertex(normals, v2normal);
        PushVertex(coords, bv1);
        PushVertex(normals, v1normal);
        PushVertex(coords, bv2);
        PushVertex(normals, v2normal);
    }
}

//TODO: if you are unhappy with how the UV map is applying the texture to the sphere
//you can tinker with it here
//I'm not happy with it either but it works well enough for now
glm::vec2 vert_to_uv(glm::vec3& vert){
    float xz_angle = atan2(vert.x, vert.z);
    float xz_dist = glm::vec2(vert.x, vert.z).length();
    float y_angle = atan2(vert.y, xz_dist);
    
    auto vec = glm::vec2(xz_angle / (2 * M_PI) + 0.5, y_angle / (2 * M_PI) + 0.5);
    return vec;
}

//APPROACH 1: SEPERATE ARRAY FOR VERTEXES, NORMALS, UVS, AND INDEXES
// FAILURE: INDEXES ARE COMPLETELY WRONG
//
// Sets up a sphere with equation x^2+y^2+z^2=1
//
void GetSphereData(std::vector<float> &coords, std::vector<float> &normals, std::vector<float> &uvs)
{
    int recursionLevel = 3;
    std::vector<Triangle> list;
    {
        Triangle t;
        t.v0 = glm::vec3(1.0f, 0.0f, 0.0f);
        t.v1 = glm::vec3(0.0f, 1.0f, 0.0f);
        t.v2 = glm::vec3(0.0f, 0.0f, 1.0f);
        list.push_back(t);
    }
    for (int r = 0; r < recursionLevel; r++)
    {
        list = SplitTriangle(list);
    }

    for (int octant = 0; octant < 8; octant++)
    {
        glm::mat4 view(1.0f);
        float angle = 90.0f * (octant % 4);
        if (angle != 0.0f)
            view = glm::rotate(view, glm::radians(angle), glm::vec3(1, 0, 0));
        if (octant >= 4)
            view = glm::rotate(view, glm::radians(180.0f), glm::vec3(0, 0, 1));
        for (int i = 0; i < list.size(); i++)
        {
            Triangle t = list[i];
            float mag_reci;
            glm::vec3 v0 = view * glm::vec4(t.v0, 1.0f);
            glm::vec3 v1 = view * glm::vec4(t.v1, 1.0f);
            glm::vec3 v2 = view * glm::vec4(t.v2, 1.0f);
            mag_reci = 1.0f / glm::length(v0);
            v0 = glm::vec3(v0.x * mag_reci, v0.y * mag_reci, v0.z * mag_reci);
            mag_reci = 1.0f / glm::length(v1);
            v1 = glm::vec3(v1.x * mag_reci, v1.y * mag_reci, v1.z * mag_reci);
            mag_reci = 1.0f / glm::length(v2);
            v2 = glm::vec3(v2.x * mag_reci, v2.y * mag_reci, v2.z * mag_reci);
            PushVertex(coords, v0);
            PushVertex(coords, v1);
            PushVertex(coords, v2);
            PushVertex(normals, v0);
            PushVertex(normals, v1);
            PushVertex(normals, v2);

            auto uv0 = vert_to_uv(v0);
            auto uv1 = vert_to_uv(v1);
            auto uv2 = vert_to_uv(v2);

            PushUV(uvs, uv0);
            PushUV(uvs, uv1);
            PushUV(uvs, uv2);
        }
    }
}


//APPROACH 2: INTERLEAVE DATA AS [v1_x, v1_y, v1_z, n1_x, n1_y, n1_z, uv1_u, uv1_v, v2_x, v2_y, ...]
//FAILURE: NO SPHERE RENDERS
//
// Sets up a sphere with equation x^2+y^2+z^2=1
//
void GetSphereData(std::vector<float> &data)
{
    int recursionLevel = 3;
    std::vector<Triangle> list;
    {
        Triangle t;
        t.v0 = glm::vec3(1.0f, 0.0f, 0.0f);
        t.v1 = glm::vec3(0.0f, 1.0f, 0.0f);
        t.v2 = glm::vec3(0.0f, 0.0f, 1.0f);
        list.push_back(t);
    }
    for (int r = 0; r < recursionLevel; r++)
    {
        list = SplitTriangle(list);
    }

    for (int octant = 0; octant < 8; octant++)
    {
        glm::mat4 view(1.0f);
        float angle = 90.0f * (octant % 4);
        if (angle != 0.0f)
            view = glm::rotate(view, glm::radians(angle), glm::vec3(1, 0, 0));
        if (octant >= 4)
            view = glm::rotate(view, glm::radians(180.0f), glm::vec3(0, 0, 1));
        for (int i = 0; i < list.size(); i++)
        {
            Triangle t = list[i];
            float mag_reci;
            glm::vec3 v0 = view * glm::vec4(t.v0, 1.0f);
            glm::vec3 v1 = view * glm::vec4(t.v1, 1.0f);
            glm::vec3 v2 = view * glm::vec4(t.v2, 1.0f);
            mag_reci = 1.0f / glm::length(v0);
            v0 = glm::vec3(v0.x * mag_reci, v0.y * mag_reci, v0.z * mag_reci);
            mag_reci = 1.0f / glm::length(v1);
            v1 = glm::vec3(v1.x * mag_reci, v1.y * mag_reci, v1.z * mag_reci);
            mag_reci = 1.0f / glm::length(v2);
            v2 = glm::vec3(v2.x * mag_reci, v2.y * mag_reci, v2.z * mag_reci);
            auto uv0 = vert_to_uv(v0);
            auto uv1 = vert_to_uv(v1);
            auto uv2 = vert_to_uv(v2);

            PushVertex(data, v0);
            PushVertex(data, v0);
            PushUV(data, uv0);

            PushVertex(data, v1);
            PushVertex(data, v1);
            PushUV(data, uv1);

            PushVertex(data, v2);
            PushVertex(data, v2);
            PushUV(data, uv2);        
        }
    }
}

//
//
// PART 2: RenderManager module
//
//

void _print_shader_info_log(GLuint shader_index)
{
    int max_length = 2048;
    int actual_length = 0;
    char shader_log[2048];
    glGetShaderInfoLog(shader_index, max_length, &actual_length, shader_log);
    printf("shader info log for GL index %u:\n%s\n", shader_index, shader_log);
}

std::string load_text(const char* filename){
    std::ifstream t(filename);
    //C++ is an affront to god, how can you possibly justify this
    std::string str((std::istreambuf_iterator<char>(t)), std::istreambuf_iterator<char>());
    return str;
}

class RenderManager
{
public:
    enum ShapeType
    {
        SPHERE,
        CYLINDER
    };

    RenderManager();
    void SetView(glm::vec3 &c, glm::vec3&, glm::vec3 &, glm::vec3 &);
    void SetUpGeometry();
    void Render(ShapeType, glm::mat4 model);
    GLFWwindow *GetWindow() { return window; };

private:
    glm::vec3 color;
    GLuint sphereVAO;
    GLuint sphereNumPrimitives;
    GLuint cylinderVAO;
    GLuint cylinderNumPrimitives;
    GLuint mvploc;
    GLuint camloc;
    GLuint ldirloc;
    GLuint lposloc;
    GLuint cdirloc;
    GLuint model_mat;
    GLuint alb_tex;
    GLuint rgh_tex;
    GLuint met_tex;
    glm::mat4 projection;
    glm::mat4 view;
    glm::vec3 light_pos;
    GLuint shaderProgram;
    GLFWwindow *window;

    void SetUpWindowAndShaders();
    void MakeModelView(glm::mat4 &);
};

RenderManager::RenderManager()
{
    SetUpWindowAndShaders();
    SetUpGeometry();
    projection = glm::perspective(
        glm::radians(45.0f), (float)1000 / (float)1000, 5.0f, 100.0f);

    // Get a handle for our MVP and color uniforms
    mvploc = glGetUniformLocation(shaderProgram, "MVP");
    //model_mat = glGetUniformLocation(shaderProgram, "model_mat");
    ldirloc = glGetUniformLocation(shaderProgram, "lightdir");
    cdirloc = glGetUniformLocation(shaderProgram, "cameradir");

    //TODO: PICK IMAGE TO LOAD HERE
    //might want to pass the base path in as an argument from main so you can change textures without recompiling
    //this is only the base path, _Roughness.jpg, _Metalness.jpg, and _Color.jpg will be appened by their loaders
    setup_opengl_images("./assets/MetalPlates006_1K", &alb_tex, &rgh_tex, &met_tex);
}

void RenderManager::SetView(glm::vec3 &camera, glm::vec3 &light, glm::vec3 &origin, glm::vec3 &up)
{
    glm::mat4 v = glm::lookAt(
        camera, // Camera in world space
        origin, // looks at the origin
        up      // and the head is up
    );
    view = v;

    //NOT NORMALIZED
    glm::vec3 camdir = camera - origin;

    //NOT NORMALIZED
    glm::vec3 lightdir = light - origin;

    glUniform3fv(ldirloc, 1, &lightdir[0]);
    glUniform3fv(cdirloc, 1, &camdir[0]);
};

void RenderManager::SetUpWindowAndShaders()
{
    // start GL context and O/S window using the GLFW helper library
    if (!glfwInit())
    {
        fprintf(stderr, "ERROR: could not start GLFW3\n");
        exit(EXIT_FAILURE);
    }

    glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 4);
    glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 0);
    glfwWindowHint(GLFW_OPENGL_FORWARD_COMPAT, GL_TRUE);
    glfwWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_CORE_PROFILE);

    window = glfwCreateWindow(1000, 1000, "PBR TIME", NULL, NULL);
    if (!window)
    {
        fprintf(stderr, "ERROR: could not open window with GLFW3\n");
        glfwTerminate();
        exit(EXIT_FAILURE);
    }
    glfwMakeContextCurrent(window);
    // start GLEW extension handler
    glewExperimental = GL_TRUE;
    glewInit();

    // get version info
    const GLubyte *renderer = glGetString(GL_RENDERER); // get renderer string
    const GLubyte *version = glGetString(GL_VERSION);   // version as a string
    printf("Renderer: %s\n", renderer);
    printf("OpenGL version supported %s\n", version);

    // tell GL to only draw onto a pixel if the shape is closer to the viewer
    glEnable(GL_DEPTH_TEST); // enable depth-testing
    glDepthFunc(GL_LESS);    // depth-testing interprets a smaller value as "closer"

    std::string vertex_shader = load_text("pbr_vert.glsl");
    std::string fragment_shader = load_text("pbr_frag.glsl");

    const GLchar* vstxt = vertex_shader.c_str();
    const GLchar* fstxt = fragment_shader.c_str();

    GLuint vs = glCreateShader(GL_VERTEX_SHADER);
    glShaderSource(vs, 1, &(vstxt), NULL);
    glCompileShader(vs);
    int params = -1;
    glGetShaderiv(vs, GL_COMPILE_STATUS, &params);
    if (GL_TRUE != params)
    {
        fprintf(stderr, "ERROR: GL shader index %i did not compile\n", vs);
        _print_shader_info_log(vs);
        exit(EXIT_FAILURE);
    }

    GLuint fs = glCreateShader(GL_FRAGMENT_SHADER);
    glShaderSource(fs, 1, &(fstxt), NULL);
    glCompileShader(fs);
    glGetShaderiv(fs, GL_COMPILE_STATUS, &params);
    if (GL_TRUE != params)
    {
        fprintf(stderr, "ERROR: GL shader index %i did not compile\n", fs);
        _print_shader_info_log(fs);
        exit(EXIT_FAILURE);
    }

    shaderProgram = glCreateProgram();
    glAttachShader(shaderProgram, fs);
    glAttachShader(shaderProgram, vs);
    glLinkProgram(shaderProgram);
    glUseProgram(shaderProgram);

    //TODO: use glUniform1i and glGetUniformLocation to attach the textures to uniform locations
    //map "alb_tex" to 0
    //map "rgh_tex" to 1
    //map "met_tex" to 2
}

void RenderManager::MakeModelView(glm::mat4 &model)
{
    glm::mat4 modelview = projection * view * model;
    glUniformMatrix4fv(model_mat, 1 ,GL_FALSE, &model[0][0]);
    glUniformMatrix4fv(mvploc, 1, GL_FALSE, &modelview[0][0]);
}

void RenderManager::Render(ShapeType st, glm::mat4 model)
{
    int numPrimitives = 0;
    if (st == SPHERE)
    {
        glActiveTexture(GL_TEXTURE0);
        glBindTexture(GL_TEXTURE_2D, alb_tex);
        glActiveTexture(GL_TEXTURE1);
        glBindTexture(GL_TEXTURE_2D, rgh_tex);
        glActiveTexture(GL_TEXTURE2);
        glBindTexture(GL_TEXTURE_2D, met_tex);
        
        glBindVertexArray(sphereVAO);
        numPrimitives = sphereNumPrimitives;
    }
    else if (st == CYLINDER)
    {
        glBindVertexArray(cylinderVAO);
        numPrimitives = cylinderNumPrimitives;
    }
    MakeModelView(model);
    glDrawElements(GL_TRIANGLES, numPrimitives, GL_UNSIGNED_INT, NULL);
}

//ONLY USED FOR CYLINDER
void SetUpVBOs(std::vector<float> &coords, std::vector<float> &normals,
               GLuint &points_vbo, GLuint &normals_vbo, GLuint &index_vbo)
{
    int numIndices = coords.size() / 3;
    std::vector<GLuint> indices(numIndices);
    for (int i = 0; i < numIndices; i++)
        indices[i] = i;

    points_vbo = 0;
    glGenBuffers(1, &points_vbo);
    glBindBuffer(GL_ARRAY_BUFFER, points_vbo);
    glBufferData(GL_ARRAY_BUFFER, coords.size() * sizeof(float), coords.data(), GL_STATIC_DRAW);

    normals_vbo = 0;
    glGenBuffers(1, &normals_vbo);
    glBindBuffer(GL_ARRAY_BUFFER, normals_vbo);
    glBufferData(GL_ARRAY_BUFFER, normals.size() * sizeof(float), normals.data(), GL_STATIC_DRAW);

    index_vbo = 0; // Index buffer object
    glGenBuffers(1, &index_vbo);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, index_vbo);
    glBufferData(GL_ELEMENT_ARRAY_BUFFER, indices.size() * sizeof(GLuint), indices.data(), GL_STATIC_DRAW);
}

//APPROACH 1
void SetUpVBOsSphere(std::vector<float> &coords, std::vector<float> &normals, std::vector<float> &uvs,
               GLuint &points_vbo, GLuint &normals_vbo, GLuint &uv_vbo, GLuint &index_vbo)
{
    int numIndices = coords.size() / 3;
    std::vector<GLuint> indices(numIndices);
    for (int i = 0; i < numIndices; i++)
        indices[i] = i;

    points_vbo = 0;
    glGenBuffers(1, &points_vbo);
    glBindBuffer(GL_ARRAY_BUFFER, points_vbo);
    glBufferData(GL_ARRAY_BUFFER, coords.size() * sizeof(float), coords.data(), GL_STATIC_DRAW);

    normals_vbo = 0;
    glGenBuffers(1, &normals_vbo);
    glBindBuffer(GL_ARRAY_BUFFER, normals_vbo);
    glBufferData(GL_ARRAY_BUFFER, normals.size() * sizeof(float), normals.data(), GL_STATIC_DRAW);

    uv_vbo = 0;
    glGenBuffers(1, &uv_vbo);
    glBindBuffer(GL_ARRAY_BUFFER, uv_vbo);
    glBufferData(GL_ARRAY_BUFFER, uvs.size() * sizeof(float), uvs.data(), GL_STATIC_DRAW);

    index_vbo = 0; // Index buffer object
    glGenBuffers(1, &index_vbo);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, index_vbo);
    glBufferData(GL_ELEMENT_ARRAY_BUFFER, indices.size() * sizeof(GLuint), indices.data(), GL_STATIC_DRAW);
}

//APPROACH 2
void SetUpVBOsSphereIL(std::vector<float>& data, GLuint &data_vbo, GLuint &index_vbo){
    int numIndices = data.size() / 8;
    std::vector<GLuint> indices(numIndices);
    for (int i = 0; i < numIndices; i++)
        indices[i] = i;

    data_vbo = 0;
    glGenBuffers(1, &data_vbo);
    glBindBuffer(GL_ARRAY_BUFFER, data_vbo);
    glBufferData(GL_ARRAY_BUFFER, data.size() * sizeof(float), data.data(), GL_STATIC_DRAW);

    index_vbo = 0; // Index buffer object
    glGenBuffers(1, &index_vbo);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, index_vbo);
    glBufferData(GL_ELEMENT_ARRAY_BUFFER, indices.size() * sizeof(GLuint), indices.data(), GL_STATIC_DRAW);
}

void RenderManager::SetUpGeometry()
{
    //APPROACH 1
    std::vector<float> sphereCoords;
    std::vector<float> sphereNormals;
    std::vector<float> sphereUVs;
    GetSphereData(sphereCoords, sphereNormals, sphereUVs);
    sphereNumPrimitives = sphereCoords.size() / 3;
    GLuint sphere_points_vbo, sphere_normals_vbo, sphere_uvs_vbo, sphere_indices_vbo;
    SetUpVBOsSphere(sphereCoords, sphereNormals, sphereUVs,
              sphere_points_vbo, sphere_normals_vbo, sphere_uvs_vbo, sphere_indices_vbo);


    std::vector<float> cylCoords;
    std::vector<float> cylNormals;
    GetCylinderData(cylCoords, cylNormals);
    cylinderNumPrimitives = cylCoords.size() / 3;
    GLuint cyl_points_vbo, cyl_normals_vbo, cyl_indices_vbo;
    SetUpVBOs(cylCoords, cylNormals,
              cyl_points_vbo, cyl_normals_vbo, cyl_indices_vbo);

    GLuint vao[2];
    glGenVertexArrays(2, vao);

    glBindVertexArray(vao[SPHERE]);
    sphereVAO = vao[SPHERE];
    //APPROACH 1
    glBindBuffer(GL_ARRAY_BUFFER, sphere_points_vbo);
    glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 0, NULL);
    glBindBuffer(GL_ARRAY_BUFFER, sphere_normals_vbo);
    glVertexAttribPointer(1, 3, GL_FLOAT, GL_FALSE, 0, NULL);
    glBindBuffer(GL_ARRAY_BUFFER, sphere_uvs_vbo);
    glVertexAttribPointer(2, 2, GL_FLOAT, GL_FALSE, 0, NULL);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, sphere_indices_vbo);
    glEnableVertexAttribArray(0);
    glEnableVertexAttribArray(1);
    glEnableVertexAttribArray(2);


    glBindVertexArray(vao[CYLINDER]);
    cylinderVAO = vao[CYLINDER];
    glBindBuffer(GL_ARRAY_BUFFER, cyl_points_vbo);
    glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 0, NULL);
    glBindBuffer(GL_ARRAY_BUFFER, cyl_normals_vbo);
    glVertexAttribPointer(1, 3, GL_FLOAT, GL_FALSE, 0, NULL);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, cyl_indices_vbo);
    glEnableVertexAttribArray(0);
    glEnableVertexAttribArray(1);
}

//
// PART3: main function
//

glm::mat4 RotateMatrix(float degrees, float x, float y, float z)
{
    glm::mat4 identity(1.0f);
    glm::mat4 rotation = glm::rotate(identity,
                                     glm::radians(degrees),
                                     glm::vec3(x, y, z));
    return rotation;
}

glm::mat4 ScaleMatrix(double x, double y, double z)
{
    glm::mat4 identity(1.0f);
    glm::vec3 scale(x, y, z);
    return glm::scale(identity, scale);
}

glm::mat4 TranslateMatrix(double x, double y, double z)
{
    glm::mat4 identity(1.0f);
    glm::vec3 translate(x, y, z);
    return glm::translate(identity, translate);
}

void set_up_sphere(RenderManager &rm, double angle)
{
    glm::mat4 identity(1.0f);
    glm::mat4 rot = RotateMatrix(angle, 0, 1, 0);
    rm.Render(RenderManager::SPHERE, identity * rot);
}



int main()
{
    RenderManager rm;
    GLFWwindow *window = rm.GetWindow();

    glm::vec3 origin(0, 0, 0);
    glm::vec3 up(0, 1, 0);

    glm::vec3 camera_pos(10, 2, 0);
    glm::vec3 light_pos(5,5,5);

    while (!glfwWindowShouldClose(window))
    {
        rm.SetView(camera_pos, light_pos, origin, up);

        // wipe the drawing surface clear
        glClearColor(0.3, 0.3, 0.8, 1.0);
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

        set_up_sphere(rm, 0);

        // update other events like input handling
        glfwPollEvents();
        // put the stuff we've been drawing onto the display
        glfwSwapBuffers(window);
    }

    // close GL context and any other GLFW resources
    glfwTerminate();
    return 0;
}